import asyncio
import logging
from typing import Optional, List, Any, Dict, Tuple
import numpy as np
from astropy.coordinates import SkyCoord
import astropy.units as u

from pyobs.events import MotionStatusChangedEvent, Event
from pyobs.mixins import FitsNamespaceMixin, MotionStatusMixin
from pyobs.modules import Module, timeout
from pyobs.utils.enums import MotionStatus
from pyobs.utils.parallel import event_wait
from pyobs.utils.threads import LockWithAbort
from pyobs.interfaces import IRotation, IMotion, IFocuser, ICalibrate, IFitsHeaderBefore, IPointingRaDec
from pyobs.utils.time import Time

from .geminidriver import GeminiDriver, GeminiCommException, Vocab

log = logging.getLogger(__name__)


class GeminiFocuserRotator(
    FitsNamespaceMixin,
    MotionStatusMixin,
    Module,
    IRotation,
    IFocuser,
    ICalibrate,
    IPointingRaDec,
    IFitsHeaderBefore,
):
    """Pyobs module for operating an Optec Inc GEMINI focuser/rotator."""

    def __init__(
        self,
        serial_config: Optional[Dict[str, Any]] = None,
        fits_config: Optional[Dict[str, Any]] = None,
        focus_offset: float = 0.0,
        rotation_offset: float = 0.0,
        *args: Any,
        **kwargs: Any,
    ):
        Module.__init__(self, *args, **kwargs)

        # add thread func
        self.add_background_task(self._gdriver_update_func, True)
        self.add_background_task(self._rotation_tracker_func, True)

        # store
        self.focus = 0.0
        self.rotation = 0.0
        self.follow = None

        # FOCUSING STUFF
        self._focus_lock = asyncio.Lock()
        self._focus_abort = asyncio.Event()
        self._focus_accur = 0.0  # MM
        self._focus_offset = focus_offset

        # ROTATOR STUFF
        self._rotation_lock = asyncio.Lock()
        self._rotation_abort = asyncio.Event()
        self._skycoord: Optional[SkyCoord] = None
        self._rotation_accur = 0.0  # DEG
        self._rotation_offset = rotation_offset

        # TEMPERATURE SENSOR
        self._T = None

        # SERIAL CONFIGURATION DICTIONARY
        if serial_config is None:
            self._serial_config = {
                "port": "/dev/ttyUSB0",
                "baudrate": 115200,
                "timeout": 0.1,
            }
        else:
            self._serial_config = serial_config

        # driver
        self._driver: Optional[GeminiDriver] = None

        # FITS HEADER CONFIGURATION
        if fits_config is None:
            self._fits_config = {
                "focus": ("GEM-FOCU", "focus of the Gemini focusser [mm]"),
                "focus-offset": ("GEM-FOFF", "constant Gemini focus offset [mm]"),
                "focus-motion": (
                    "GEM-FMOT",
                    "motion status of the Gemini focusser [mm]",
                ),
                "rotation": ("GEM-ROTA", "angle of the Gemini rotator [mm]"),
                "rotation-offset": ("GEM-ROFF", "constant Gemini rotation offset [mm]"),
                "rotation-motion": (
                    "GEM-RMOT",
                    "motion status of the Gemini rotator [mm]",
                ),
                "temperature": ("GEM-TEMP", "temperature of the Gemini sensor [C]"),
            }
        else:
            self._fits_config = fits_config

        # mixins
        FitsNamespaceMixin.__init__(self, *args, **kwargs)
        MotionStatusMixin.__init__(self, **kwargs, motion_status_interfaces=["IFocuser", "IRotation"])

    async def open(self) -> None:
        """Open module."""
        await Module.open(self)

        # subscribe to events
        if self.comm:
            if self.follow:
                # self.comm.register_event(TelescopeMovingEvent, self._telescope_event)
                self.comm.register_event(MotionStatusChangedEvent, self._telescope_event)

        # create driver and open it
        self._driver = GeminiDriver(**self._serial_config)

        # calibrate
        log.info("Calibrating unit...")
        await self._driver.calibrate()
        self._focus_accur = self._driver.get_focus_accuracy()
        self._rotation_accur = self._driver.get_rotation_accuracy()

        # open mixins
        await MotionStatusMixin.open(self)

    async def close(self) -> None:
        """Close module."""
        await Module.close(self)

        log.info("Closing hardware connection...")
        if self._driver is not None:
            self._driver = None

    @timeout(600000)
    async def calibrate(self, **kwargs: Any) -> None:
        """Calibrate the device."""
        if self._driver is None:
            return

        # reset
        self._skycoord = None

        # acquire focus lock
        async with LockWithAbort(self._focus_lock, self._focus_abort):
            # acquire rotation lock
            async with LockWithAbort(self._rotation_lock, self._rotation_abort):
                # start homing for both
                log.info("Start homing focus...")
                if not self._driver.start_home_focus():
                    raise ValueError("Could not start homing for focus.")
                log.info("Start homing rotation...")
                if not self._driver.start_home_rotation():
                    raise ValueError("Could not start homing for rotation.")

                # wait for both
                while True:
                    # both homed?
                    if self._driver.focus_is_homed() and self._driver.rotation_is_homed():
                        # finished
                        log.info("Homing successful.")
                        return

                    # abort any?
                    if self._focus_abort.is_set() or self._rotation_abort.is_set():
                        log.warning("Homing aborted.")
                        return

                    # sleep a little (can only wait on unset events
                    if not self._focus_abort.is_set():
                        await event_wait(self._focus_abort, 1)
                    else:
                        await event_wait(self._rotation_abort, 1)

    async def _gdriver_update_func(self) -> None:
        log.info("Starting GEMINI driver update thread...")

        while True:
            # do update
            await self._update_status()

            # sleep a little
            await asyncio.sleep(1)

    async def _update_status(self) -> None:
        if self._driver is not None:
            # get data
            fdict = await self._driver.get_focus_status()
            rdict = await self._driver.get_rotation_status()

            # get current focus and rotation
            self.focus = fdict.data[Vocab.FOCUS_MM]
            self.rotation = rdict.data[Vocab.POSANG_DEG]

            # get temp
            # TODO: find out
            # self._T = fdict.response[Vocab.CURRENT_TEMP.value]

            # get motion status
            await self._change_motion_status(self._motion_status(fdict.response), interface="IFocuser")
            await self._change_motion_status(self._motion_status(rdict.response), interface="IRotation")

    def _motion_status(self, stat: Dict[str, Any]) -> MotionStatus:
        """
        Extracts the IMotion status from a dictionary returned
        by the driver's status method.
        """
        if "IsHoming" in stat and stat["IsHoming"]:
            return MotionStatus.INITIALIZING
        if "IsMoving" in stat and stat["IsMoving"]:
            return MotionStatus.SLEWING
        if self._skycoord is None:
            return MotionStatus.IDLE
        else:
            return MotionStatus.TRACKING

    @timeout(300000)
    async def set_focus(self, focus: float, **kwargs: Any) -> None:
        """Sets new focus.

        Args:
            focus: New focus value.

        Raises:
            MoveError: If telescope cannot be moved.
            InterruptedError: If movement was aborted.
        """
        if self._driver is None:
            return

        # acquire lock
        async with LockWithAbort(self._focus_lock, self._focus_abort):
            # set focus value
            try:
                log.info("Setting focus to %.2f...", focus)
                await self._driver.set_focus(focus)
            except GeminiCommException:
                log.exception("Could not set new focus.")

            # sleep a little and force update
            await event_wait(self._focus_abort, 1)
            await self._update_status()

            while not self._focus_abort.is_set() and await self.get_motion_status("IFocuser") == MotionStatus.SLEWING:
                # sleep a little
                await event_wait(self._focus_abort, 1)

            # aborted?
            if self._focus_abort.is_set():
                raise InterruptedError("Setting focus was interrupted.")

        # success
        log.info("Successfully set new focus.")

    async def get_focus(self, **kwargs: Any) -> float:
        """Return current focus.

        Returns:
            Current focus.
        """
        return self.focus

    @timeout(300000)
    async def set_rotation(self, angle: float, **kwargs: Any) -> None:
        """Sets the rotation angle to the given value in degrees."""
        if self._driver is None:
            return

        # acquire lock
        async with LockWithAbort(self._rotation_lock, self._rotation_abort):
            # set focus value
            try:
                log.info("Setting rotation to %.2f...", angle)
                await self._driver.set_rotation(angle)
            except GeminiCommException:
                log.exception("Could not set new rotation.")

            # sleep a little and force update
            await event_wait(self._rotation_abort, 1)
            await self._update_status()

            while (
                not self._rotation_abort.is_set() and await self.get_motion_status("IRotation") == MotionStatus.SLEWING
            ):
                # sleep a little
                await event_wait(self._rotation_abort, 1)

            # aborted?
            if self._rotation_abort.is_set():
                raise InterruptedError("Setting rotation was interrupted.")

        # success
        log.info("Successfully set new rotation.")

    async def get_rotation(self, **kwargs: Any) -> float:
        """Returns the current rotation angle."""
        return await self._driver.get_rotation()

    async def stop_motion(self, device: Optional[str] = None, **kwargs: Any) -> None:
        """Stop the motion.

        Args:
            device: Name of device to stop, or None for all.
        """
        if device is None or device == "IRotation":
            # need to stop tracking?
            if self._skycoord is not None:
                self._skycoord = None
                log.info("Stopped parallactic angle tracking.")

    async def move_radec(self, ra: float, dec: float, **kwargs: Any) -> None:
        """Tracks the position angle of a rotator for an alt-az telescope."""
        if self._driver is None:
            return

        # first, reset tracking
        self._skycoord = None

        # valid coordinates?
        if ra < 0.0 or ra > 360.0 or np.abs(dec) >= 90.0:
            raise ValueError("RA, Dec out of limits (%.2f, %.2f)." % (ra, dec))
        skycoord = SkyCoord(ra * u.deg, dec * u.deg, frame="icrs")

        # get parallactic angle
        pa = self.observer.parallactic_angle(Time.now(), skycoord).degree

        # initial rotation
        await self.set_rotation(pa + self._rotation_offset)

        # start tracking and log it
        self._skycoord = skycoord
        log.info(
            "Started target tracking of parallactic angle at %s...",
            self._skycoord.to_string(),
        )

    async def get_radec(self, **kwargs: Any) -> Tuple[float, float]:
        return 0.0, 0.0

    async def _rotation_tracker_func(self) -> None:
        if self._driver is None:
            return

        # log
        log.info("Starting rotation tracking thread...")

        while True:
            # do we have a sky coord to track?
            if self._skycoord is not None:
                # get parallactic angle
                pa = self.observer.parallactic_angle(Time.now(), self._skycoord).degree

                # get rotation
                rot = await self.get_rotation()

                # need to rotate?
                if np.abs(pa - rot) > self._rotation_accur:
                    # rotate
                    await self._driver.set_rotation(pa + self._rotation_offset)

            # sleep a little
            await asyncio.sleep(1)

    async def get_fits_header_before(
        self, namespaces: Optional[List[str]] = None, **kwargs: Any
    ) -> Dict[str, Tuple[Any, str]]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """
        hdr = {}

        # SET FOCUS HEADERS
        if "focus" in self._fits_config:
            key, comment = self._fits_config["focus"]
            hdr[key] = self.focus, comment
        if "focus-motion" in self._fits_config:
            key, comment = self._fits_config["focus-motion"]
            hdr[key] = (await self.get_motion_status("IFocuser")).value, comment
        if "focus-offset" in self._fits_config:
            key, comment = self._fits_config["focus-offset"]
            hdr[key] = self._focus_offset, comment

        # SET ROTATION HEADERS
        if "rotation" in self._fits_config:
            key, comment = self._fits_config["rotation"]
            hdr[key] = self.rotation, comment
        if "rotation-motion" in self._fits_config:
            key, comment = self._fits_config["rotation-motion"]
            hdr[key] = (await self.get_motion_status("IRotation")).value, comment
        if "rotation-offset" in self._fits_config:
            key, comment = self._fits_config["rotation-offset"]
            hdr[key] = self._rotation_offset, comment

        # TEMPERATURE SENSOR
        if "temperature" in self._fits_config:
            key, comment = self._fits_config["temperature"]
            hdr[key] = self._T, comment

        # return it
        return self._filter_fits_namespace(hdr, namespaces=namespaces, **kwargs)

    async def set_focus_offset(self, offset: float, **kwargs: Any) -> None:
        """Sets focus offset.

        Args:
            offset: New focus offset.

        Raises:
            ValueError: If given value is invalid.
            MoveError: If telescope cannot be moved.
        """
        pass

    async def get_focus_offset(self, **kwargs: Any) -> float:
        """Return current focus offset.

        Returns:
            Current focus offset.
        """
        return 0.0

    async def init(self, **kwargs: Any) -> None:
        """Initialize device.

        Raises:
            InitError: If device could not be initialized.
        """
        pass

    async def park(self, **kwargs: Any) -> None:
        """Park device.

        Raises:
            ParkError: If device could not be parked.
        """
        pass

    async def is_ready(self, **kwargs: Any) -> bool:
        """Returns the device is "ready", whatever that means for the specific device.

        Returns:
            Whether device is ready
        """
        return True

    async def _telescope_event(self, ev: Event, sender: str) -> bool:
        """Moving events from telescope.

        Args:
            event: Either a MotionStatusChangedEvent or a TelescopeMovingEvent.
            sender: Who sent it.
        """

        # first check sender against self.follow
        if self.follow is None or self.follow != sender:
            return False

        # what kind of event is it?
        if isinstance(ev, TelescopeMovingEvent):
            # do we have RA/Dec?
            if ev.ra is not None and ev.dec is not None:
                # start tracking!
                log.info("Received an event that telescope is about to start tracking. Following it... ")
                await self.track(ev.ra, ev.dec)

            else:
                # presumably slewing to fixed coordinates
                # are we currently tracking?
                if self._skycoord is not None:
                    log.info("Received event that telescope is not tracking anymore, stopping derotator movement...")
                    await self.stop_motion("IRotation")

        elif isinstance(ev, MotionStatusChangedEvent):
            # we want the ITelescope event and anything except TRACKING and SLEWING (might end up in race condition)
            if "ITelescope" in ev.interfaces and ev.interfaces["ITelescope"] not in [
                MotionStatus.TRACKING.value,
                MotionStatus.SLEWING.value,
            ]:
                # are we currently tracking?
                if self._skycoord is not None:
                    # stop it
                    log.info("Received event that telescope is not tracking anymore, stopping derotator movement...")
                    await self.stop_motion("IRotation")


__all__ = ["GeminiFocuserRotator"]



"""
Contains the OPTEC API for the GEMINI focusser/rotator, Rev. 2.1

The GEMINI commands have the following syntax:

    <{F|R|H}{d}{ii}{cmd}{data}>

where
    F 	:	focusser
    R	:	rotator
    H	:	hub
    {d}	:	device ID, must be = 1 for GEMINI !!!
    {ii}	:	transaction ID from 01 to 99
    {cmd}	:	command
   {data}	:	option input data field

The GEMINI responses have the syntax:

   !{ii}
   {response}
   {END|SET}

"""
from typing import Dict, Any, Tuple, List

# A COMMAND CONSISTS OF name,device,command_string,input_format,output_format,end_string,help
gemini_commands: Dict[str, Dict[str, Any]] = {
    "CENTER": {
        "cmd": "center",
        "device": "F",
        "format": None,
        "formats": None,
        "end": "END",
        "help": "center the focuser",
    },
    "DOHALT": {
        "cmd": "halt",
        "device": "FR",
        "format": None,
        "formats": None,
        "end": "END",
        "help": "immediately stop motion",
    },
    "DOHOME": {
        "cmd": "home",
        "device": "FR",
        "format": None,
        "formats": None,
        "end": "END",
        "help": "move to home position",
    },
    "GETCFG": {
        "cmd": "get configuration settings",
        "device": "FRH",
        "format": None,
        "formats": [
            "BLCompON = {0:d}",
            "BLCSteps = {0:d}",
            "Dev Type = {0}",
            "HOnStart = {0:d}",
            "MaxSteps = {0:d}",
            "Nickname = {0}",
            "CurrenTC = {0}",
            "TC Start = {0:d}",
            "TCMode A = {0:d}",
            "TCMode B = {0:d}",
            "TCMode C = {0:d}",
            "TCMode D = {0:d}",
            "TCMode E = {0:d}",
            "TComp On = {0:d}",
            "iReverse = {0:d}",
            "MaxSpeed = {0:d}",
            "PAOffset = {0:d}",
            "Firmware = {0}",
            "LEDBrite = {0:d}",
            "HandCtrl = {0:d}",
            "Wired IP = {0}",
            "WiFi Mod = {0:d}",
            "WiFiConn = {0:d}",
            "WiFiFVOK = {0:d}",
            "WiFiFirm = {0}",
            "WiFiSSID = {0}",
            "WiFiAddr = {0}",
            "WiFiSecM = {0}",
            "WiFiSecK = {0}",
        ],
        "end": "END",
        "help": "get the configuration parameters (6 common, 8 focus, 3 rotator, 12 hub)",
    },
    "GETDNN": {
        "cmd": "get device nickname",
        "device": "FR",
        "format": None,
        "formats": ["Nickname = {0}"],
        "end": "END",
        "help": "get the nickname of the device",
    },
    "GETSTA": {
        "cmd": "get status",
        "device": "FRH",
        "format": None,
        "formats": [
            "CurentPA = {0:0d}",
            "CurrStep = {0:0d}",
            "CurrTemp = {0:f}",
            "Is Homed = {0:0d}",
            "IsHoming = {0:0d}",
            "IsMoving = {0:0d}",
            "TargetPA = {0:0d}",
            "TargStep = {0:0d}",
            "TempProb = {0:0d}",
        ],
        "end": "END",
        "help": "get the device status",
    },
    "MOVABS": {
        "cmd": "move absolute position",
        "device": "F",
        "format": "{0:0d}",
        "formats": None,
        "end": "END",
        "help": "move focuser to an absolute position in encoder steps",
    },
    "MOVEPA": {
        "cmd": "move absolute position angle",
        "device": "R",
        "format": "{0:06d}",
        "formats": None,
        "end": "END",
        "help": "rotate to an absolute position in milli-degrees (0 to 359999)",
    },
    "REBOOT": {
        "cmd": "reboot",
        "device": "H",
        "format": None,
        "formats": None,
        "end": "SET",
        "help": "soft reboot of hub",
    },
    "RESETH": {
        "cmd": "reset to factory defaults",
        "device": "H",
        "format": None,
        "formats": None,
        "end": "SET",
        "help": "reset to factory default settings",
    },
    "SETBCE": {
        "cmd": "set backlash compensation enabled",
        "device": "FR",
        "format": "{0:d}",
        "formats": None,
        "end": "SET",
        "help": "set whether backlash compensation is enabled (0=no,1=yes)",
    },
    "SETBCS": {
        "cmd": "set backlash compensation steps",
        "device": "FR",
        "format": "{0:0d}",
        "formats": None,
        "end": "SET",
        "help": "sets the number of steps used for backlash compensation (0-99)",
    },
    "SETHOS": {
        "cmd": "set home on start",
        "device": "FR",
        "format": "{0:d}",
        "formats": None,
        "end": "SET",
        "help": "set whether the device is homed upon starting",
    },
    "SETLED": {
        "cmd": "set LED brightness",
        "device": "H",
        "format": "{0:02d}",
        "formats": None,
        "end": "SET",
        "help": "sets the hub LED brightness in percent (0=off)",
    },
    "SETDNN": {
        "cmd": "set device nickname",
        "device": "FR",
        "format": "{0:16s}",
        "formats": None,
        "end": "SET",
        "help": "set the device nickname",
    },
    "SETTCE": {
        "cmd": "set temperature compensation enabled",
        "device": "F",
        "format": "{0:d}",
        "formats": None,
        "end": "SET",
        "help": "set whether the temparature compensation should be used (0=no,1=yes)",
    },
    "SETTCM": {
        "cmd": "set active temperature compensation mode",
        "device": "F",
        "format": "{0}",
        "formats": None,
        "end": "SET",
        "help": "set the temparature compensation model (A|B|C|D|E)",
    },
    "SETTCC": {
        "cmd": "set temperature compensation coefficient",
        "device": "F",
        "format": "{0}{1:+04d}",
        "formats": None,
        "end": "SET",
        "help": "set the temparature compensation coefficent by #,value",
    },
    "SETTCS": {
        "cmd": "set temperature compensation at start enabled",
        "device": "F",
        "format": "{0:d}",
        "formats": None,
        "end": "SET",
        "help": "set whether to use temparature compensation at start",
    },
    "SETREV": {
        "cmd": "set reverse property",
        "device": "R",
        "format": "{0:d}",
        "formats": None,
        "end": "SET",
        "help": "set whether to reverse all angles",
    },
}


# NOTE: ALL INTERNAL AND WIFI COMMANDS MISSING FROM THIS LIST!


def gemini_cmd(cmd: str, *args: Any, dev: str = "", trans_id: int = 0) -> str:
    """
    Constructs a GEMINI command for the device 'F' (focuser),
    'R' (rotator), or 'H' (hub) for an assumed device ID=1, the
    transaction ID, and whatever input args are needed.
    """
    if cmd not in gemini_commands:
        raise ValueError("{0} is not a GEMINI command!".format(cmd))
    info = gemini_commands[cmd]
    if dev not in info["device"]:
        raise ValueError("{0} does not match GEMINI device {1}".format(dev, info["device"]))

    s1 = "<{0}1{1:02d}".format(dev, trans_id)
    s2 = cmd
    if info["format"] is None:
        s3 = ""
    else:
        try:
            s3 = info["format"].format(*args)
        except:
            raise ValueError("format {0} does not work for {1}".format(info["format"], cmd))
    s4 = ">"
    return s1 + s2 + s3 + s4


def gemini_parse_output(cmd: str, raw_response: List[bytes]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Parse the raw GEMINI output, a list of bytearray's from serial.readline,
    into a dictionary.
    Raises ValueError if the response is an error message or unparseable.
    """

    # check
    if cmd not in gemini_commands:
        raise ValueError(f"{cmd} is not a GEMINI command!")
    info = gemini_commands[cmd]
    if "end" not in info:
        raise IndexError("ending is not in info!")
    results: Dict[str, Any] = {}

    # decode and remove newlines
    try:
        rsp = [r.decode("utf8").replace("\\n", "") for r in raw_response]
    except UnicodeDecodeError:
        raise ValueError("{0!r} cannot be decoded".format(raw_response))

    # check for syntax
    if len(rsp) == 0 or not rsp[0].strip().startswith("!"):
        raise ValueError("unable to parse GEMINI response " + str(raw_response))

    # get transaction ID
    try:
        trans_id = int(rsp[0].strip()[1:])
    except:
        trans_id = None
    results["trans_id"] = trans_id
    errors = {}

    # get any output fields
    if info["formats"] is not None:
        # loop lines of output
        for r in rsp[1:-1]:
            rkey, rdata = r.split("=")
            if rkey.startswith("ERROR"):
                err = rkey[6:].strip()
                errors[err] = rdata.strip()
            else:
                # loop expected lines of output
                for f in info["formats"]:
                    pkey, pfmt = f.split("=")
                    if pkey == rkey:
                        key = pkey.strip()
                        try:
                            if "f" in pfmt:
                                # float
                                results[key] = float(rdata)
                            elif "d" in pfmt:
                                # integer
                                results[key] = int(rdata)
                            else:
                                # string
                                results[key] = rdata.strip()
                            break
                        except ValueError:
                            errors[key] = rdata
    return results, errors



import asyncio
import enum
import datetime
import logging
from typing import Optional, Any, Dict, List, cast
import numpy as np
import aioserial
from itertools import chain

from pyobs_gemini.api import gemini_commands, gemini_cmd, gemini_parse_output

log = logging.getLogger(__name__)


class GeminiTransaction:
    def __init__(
        self,
        tid: int = 0,
        cmd: Optional[str] = None,
        raw_cmd: Optional[str] = None,
        response: Optional[Dict[str, Any]] = None,
        errors: Optional[Dict[str, Any]] = None,
    ):
        """Constructor for a gemini transaction dictionary"""
        self.transaction = {"id": tid, "cmd": cmd, "raw_cmd": raw_cmd}
        self.response = {} if response is None else response
        self.errors = {} if errors is None else errors
        self.time = datetime.datetime.now()
        self.data: Dict[Vocab, Any] = {}


def has_transaction_error(d: Optional[GeminiTransaction], keys: Optional[List[str]] = None) -> bool:
    """Is there a problem with this transaction dictionary?"""
    if d is None:
        return True
    if keys is None:
        return len(d.errors) > 0
    else:
        for key in keys:
            if key in d.errors:
                return True
        return False


def dict_union(*args: Any) -> Dict[str, Any]:
    return dict(chain.from_iterable(d.items() for d in args if d is not None))


class GeminiCommException(Exception):
    pass


class Vocab(enum.Enum):
    """
    Vocabulary used for YAML configuration and status keywords.
    CURRENT_* are internal Optec keywords.
    FOCUS_MM and POSANG_DEG are calibrated quantity keys.

    YAML syntax for configuration dictionaries:
        FOCUS | ROTATION :
            HOME : bool
            INITIAL : float
            OFFSET : float
        SERIAL :
            PORT : str
            BAUDRATE : int
            TIMEOUT : float
    """

    FOCUS = "focus"
    ROTATION = "rotation"
    HUB = "hub"
    HOME = "home"
    INITIAL = "initial"
    OFFSET = "offset"
    SERIAL = "serial"
    PORT = "port"
    BAUDRATE = "baudrate"
    TIMEOUT = "timeout"
    CURRENT_TEMP = "CurrTemp"
    CURRENT_PA = "CurentPA"
    CURRENT_STEP = "CurrStep"
    FOCUS_MM = "focus_mm"
    POSANG_DEG = "pa_deg"
    MAX_STEPS = "MaxSteps"


class GeminiDriver:
    """
    Simple driver for operating the GEMINI (one needs a central driver to
    handle the serial I/O and the transaction ID's).
    """

    def __init__(self, port: str = "/dev/ttyUSB0", baudrate: int = 115200, timeout: float = 0.2):
        # lock
        self._lock = asyncio.Lock()

        # init
        self._serial: Optional[aioserial.AioSerial] = None
        self.transID = 0
        self.transactions = np.array([None for i in range(99)], dtype=dict)  # CIRCULAR BUFFER
        self.focus_model: Optional[Dict[str, float]] = None
        self.rotation_model: Optional[Dict[str, float]] = None
        self.focus_offset = 0.0
        self.rotation_offset = 0.0

        # serial connection
        self.serial = aioserial.AioSerial(port=port, timeout=timeout, baudrate=baudrate)

    async def _send_to_gemini(self, dev: str, cmd: str, *args: Any) -> GeminiTransaction:
        """
        Send the GEMINI device dev="F"|"R" the command "cmd", using any
        optional arguments.

        RETURN:
            Returns a GeminiTransaction dictionary.
        """

        # acquire lock
        async with self._lock:
            # increase ID
            self.transID += 1
            if self.transID > 99:
                self.transID = 1  # API USES 1-99

            # proper command?
            if cmd not in gemini_commands:
                return GeminiTransaction(tid=self.transID, cmd=cmd, errors={"unknown command": cmd})

            # get command format
            fmt = gemini_commands[cmd]["format"]

            # construct command
            try:
                if fmt is not None:
                    rcmd = gemini_cmd(cmd, *args, dev=dev, trans_id=self.transID)
                else:
                    rcmd = gemini_cmd(cmd, dev=dev, trans_id=self.transID)
            except ValueError as e:
                return GeminiTransaction(tid=self.transID, cmd=cmd, errors={"command error": str(e)})

            # send command
            await self.serial.write_async(bytes(rcmd, "utf8"))

            # get response
            outp = await self.serial.readlines_async()
            response, errors = gemini_parse_output(cmd, outp)

            # info
            d = GeminiTransaction(
                tid=self.transID,
                cmd=cmd,
                raw_cmd=rcmd,
                response=response,
                errors=errors,
            )

            # add to list of transactions
            self.transactions = np.roll(self.transactions, 1)
            self.transactions[0] = d

            # return resulting transaction
            return d

    def _steps_to_mm(self, steps: int) -> float:
        """Calculate the focus [mm] from the internal steps"""
        f: float = np.nan
        if self.focus_model is not None:
            f = self.focus_model["offset"] + self.focus_model["scale"] * steps
            f = max(self.focus_model["min"], min(self.focus_model["max"], f))
        return f

    def _mm_to_steps(self, focus: float) -> int:
        """Calculate the focus in steps for a given distance [mm]"""
        if self.focus_model is None:
            s = -999999
        else:
            s = int((focus - self.focus_model["offset"]) / self.focus_model["scale"])
            s = max(0, min(int(self.focus_model["max_steps"]), s))
        return s

    def _intern_to_extern(self, pa_int: int) -> float:
        """
        Calculate the external position angle [deg]
        for a given internal position angle [millideg]
        """
        m = self.rotation_model
        pa_ext: float = np.nan
        if m is not None:
            pa_ext = m["offset"] + m["scale"] * pa_int + self.rotation_offset
        return pa_ext

    def _extern_to_intern(self, pa_ext: float) -> int:
        """Calculate the internal position angle [millideg] for a given external position angle [deg]"""
        m = self.rotation_model
        if m is None:
            pa_int = -999999
        else:
            pa_int = int((pa_ext - self.rotation_offset - m["offset"]) / m["scale"])
            if pa_int < 0:
                pa_int += 360000
            if pa_int > 360000:
                pa_int -= 360000
        return pa_int

    async def calibrate(self) -> bool:
        """Calibrates the focus and rotation scales from steps to mm/deg."""

        # get focuser configuration
        cfg = await self.get_focus_config()
        response = cfg.response
        if has_transaction_error(cfg, keys=[Vocab.MAX_STEPS.value]):
            logging.error("could not get focus MaxSteps", str(cfg.errors))
            return False
        if Vocab.MAX_STEPS.value not in response:
            logging.error('"MaxSteps" not found in focus configuration!')
            return False

        # conversion between hardware and mm focus: mm = offset + scale * steps
        min_fsteps = 0
        max_fsteps = response[Vocab.MAX_STEPS.value]
        frange = 60.0  # mm
        fscale = frange / (max_fsteps - min_fsteps)
        foffset = 0.0 - fscale * 0.5 * (min_fsteps + max_fsteps)
        self.focus_model = {
            "offset": foffset,
            "scale": fscale,
            "min": -0.5 * frange,
            "max": +0.5 * frange,
            "accur": fscale * 2,
            "max_steps": max_fsteps,
        }

        # conversion between hardware and position angle: deg = offset + scale * millideg
        # 	native is 0-360000 milli-degrees, convert to -180 to +180 degrees
        min_rsteps = 0
        max_rsteps = 360000
        rrange = 360.0  # DEG
        self.rotation_model = {
            "offset": -180.0,
            "scale": 0.001,
            "min": -180.0,
            "max": +180.0,
            "accur": 0.002,
            "max_steps": max_rsteps,
        }

        return True

    async def halt_focus(self) -> bool:
        """Stops the focus motion."""
        try:
            await self._send_to_gemini("F", "DOHALT")
        except GeminiCommException as e:
            log.error(str(e))
            return False
        return True

    async def halt_rotation(self) -> bool:
        """Stops the rotation motion."""
        try:
            await self._send_to_gemini("R", "DOHALT")
        except GeminiCommException as e:
            log.error(str(e))
            return False
        return True

    async def _get_status(self, dev: str = "F") -> GeminiTransaction:
        """Gets status in the form of a transaction dictionary."""
        return await self._send_to_gemini(dev, "GETSTA")

    async def get_focus_status(self) -> GeminiTransaction:
        """
        Gets status of the focuser in the form of a transaction dictionary.
        In addtion, the calibrated focus in mm is added.
        """
        try:
            finfo = await self._get_status("F")
        except GeminiCommException as e:
            log.error(str(e))
            return GeminiTransaction(
                tid=self.transID,
                cmd="get_focus_status",
                errors={"GeminiCommException": str(e)},
            )

        if not has_transaction_error(finfo, keys=[Vocab.CURRENT_STEP.value]):
            focus_mm = self._steps_to_mm(finfo.response[Vocab.CURRENT_STEP.value])
            finfo.data[Vocab.FOCUS_MM] = focus_mm

        return finfo

    async def get_rotation_status(self) -> GeminiTransaction:
        """
        Gets status of the rotator in the form of a transaction dictionary.
        In addtion, the calibrated parallactic rotation angle in deg is added.
        """
        try:
            rinfo = await self._get_status("R")
        except GeminiCommException as e:
            log.error(str(e))
            return GeminiTransaction(
                tid=self.transID,
                cmd="get_rotation_status",
                errors={"GeminiCommException": str(e)},
            )

        if not has_transaction_error(rinfo, keys=[Vocab.CURRENT_PA.value]):
            pos_int = rinfo.response[Vocab.CURRENT_PA.value]
            pos_ext = self._intern_to_extern(pos_int)
            rinfo.data[Vocab.POSANG_DEG] = pos_ext

        return rinfo

    async def get_hub_status(self) -> GeminiTransaction:
        """Gets status of the GEMINI hub in the form of a transaction dictionary."""
        try:
            hinfo = await self._get_status("H")
        except GeminiCommException as e:
            log.error(str(e))
            return GeminiTransaction(
                tid=self.transID,
                cmd="get_rotation_status",
                errors={"GeminiCommException": str(e)},
            )

        return hinfo

    def get_focus_accuracy(self) -> float:
        if self.focus_model is None:
            return 0.0
        else:
            return self.focus_model["accur"]

    def get_rotation_accuracy(self) -> float:
        if self.rotation_model is None:
            return 0.0
        else:
            return self.rotation_model["accur"]

    async def _get_config(self, dev: str = "F") -> GeminiTransaction:
        """Gets configuration of the device in the form of a transaction dictionary."""
        return await self._send_to_gemini(dev, "GETCFG")

    async def get_focus_config(self) -> GeminiTransaction:
        """Gets configuration of the focuser in the form of a transaction dictionary."""
        try:
            return await self._get_config("F")
        except GeminiCommException as e:
            log.error(str(e))
            return GeminiTransaction(
                tid=self.transID,
                cmd="get_focus_config",
                errors={"GeminiCommException": str(e)},
            )

    async def get_rotation_config(self) -> GeminiTransaction:
        """Gets configuration of the rotator in the form of a transaction dictionary."""
        try:
            return await self._get_config("R")
        except GeminiCommException as e:
            log.error(str(e))
            return GeminiTransaction(
                tid=self.transID,
                cmd="get_focus_config",
                errors={"GeminiCommException": str(e)},
            )

    async def get_hub_config(self) -> GeminiTransaction:
        """Gets configuration of the hub in the form of a transaction dictionary."""
        try:
            return await self._get_config("H")
        except GeminiCommException as e:
            log.error(str(e))
            return GeminiTransaction(
                tid=self.transID,
                cmd="get_hub_config",
                errors={"GeminiCommException": str(e)},
            )

    async def start_home_focus(self) -> bool:
        """Starts the homing of the focuser, returns a transaction dictionary."""
        try:
            await self._send_to_gemini("F", "DOHOME")
            return True
        except GeminiCommException as e:
            log.error("Could not home focus: " + str(e))
            return False

    async def focus_is_homed(self) -> bool:
        """Whether focus is homed or not."""

        # get current status and check
        result = await self.get_focus_status()
        if has_transaction_error(result, keys=["Is Homed"]):
            return False
        elif "Is Homed" not in result.response:
            log.error('"Is Homed" not in response')
            return False
        else:
            return cast(bool, result.response["Is Homed"])

    async def start_home_rotation(self) -> bool:
        """Starts the homing of the rotator, returns a transation dictionary."""
        try:
            log.info("Homing rotation...")
            await self._send_to_gemini("R", "DOHOME")
            return True
        except GeminiCommException as e:
            log.error("Could not home rotation: " + str(e))
            return False

    async def rotation_is_homed(self) -> bool:
        """Whether rotation is homed or not."""

        # get current status and check
        result = await self.get_rotation_status()
        if has_transaction_error(result, keys=["Is Homed", "GeminiCommException"]):
            return False
        elif "Is Homed" not in result.response:
            log.error('"Is Homed" not in response')
            return False
        else:
            return cast(bool, result.response["Is Homed"])

    async def set_led(self, level: int) -> bool:
        """Sets the brightness of the hub's LED."""
        level = max(0, min(99, level))
        try:
            await self._send_to_gemini("H", "SETLED", level)
            return True
        except GeminiCommException as e:
            log.error(str(e))
            return False

    async def reboot_hub(self) -> bool:
        try:
            await self._send_to_gemini("H", "REBOOT")
            return True
        except GeminiCommException as e:
            log.error(str(e))
            return False

    async def reverse_rotation(self, flag: bool = False) -> bool:
        try:
            await self._send_to_gemini("R", "SETREV", int(flag))
            return True
        except GeminiCommException as e:
            log.error(str(e))
            return False

    async def set_focus(self, focus: float) -> bool:
        """Set focus to the given value in mm."""
        # CALCULATE INTERNAL FOCUS VALUE
        steps = self._mm_to_steps(focus)

        # SEND COMMAND TO FOCUS
        try:
            await self._send_to_gemini("F", "MOVABS", steps)
        except GeminiCommException as e:
            log.error(str(e))
            return False
        return True

    async def get_focus(self) -> float:
        """Get present focus value in mm."""
        try:
            result = await self._send_to_gemini("F", "GETSTA")
        except GeminiCommException as e:
            log.error(str(e))
            return float(np.nan)
        if Vocab.CURRENT_STEP.value in result.response:
            f = self._steps_to_mm(result.response[Vocab.CURRENT_STEP.value])
        else:
            f = np.nan
        return f

    async def set_rotation(self, pa_ext: float) -> bool:
        """Set the external position angle to the given value in degrees."""
        # CALCULATE INTERNAL POSITION ANGLE
        pa_int = self._extern_to_intern(pa_ext)

        # SEND COMMAND TO FOCUSER
        try:
            result = await self._send_to_gemini("R", "MOVEPA", pa_int)
        except GeminiCommException as e:
            log.error(str(e))
            return False
        if has_transaction_error(result):
            log.error("error while moving rotation")
            return False
        return True

    async def get_rotation(self) -> float:
        """Get present calibrated position angle [deg]."""
        try:
            result = await self._send_to_gemini("R", "GETSTA")
        except GeminiCommException as e:
            log.error(str(e))
            return False
        response = result.response
        if Vocab.CURRENT_PA.value in response:
            pa_ext = self._intern_to_extern(response["CurentPA"])
        else:
            pa_ext = np.nan
        return pa_ext

    async def get_temperature(self) -> float:
        """Get internal temperature in deg C."""
        result = await self.get_focus_status()
        response = result.response
        if "CurrTemp" in response:
            return cast(float, response["CurrTemp"])
        else:
            return float(np.nan)


__all__ = ["Vocab", "GeminiCommException", "GeminiDriver"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyobs_gemini']

package_data = \
{'': ['*']}

install_requires = \
['aioserial>=1.3,<2.0', 'astropy>=5.1,<6.0', 'pyobs-core>=1.0,<2.0']

setup_kwargs = {
    'name': 'pyobs-gemini',
    'version': '1.0.0',
    'description': 'pyobs module for Optec Gemini focusser/rotator',
    'long_description': 'None',
    'author': 'Tim-Oliver Husser',
    'author_email': 'thusser@uni-goettingen.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)
